import logging
import os
from julia_project import JuliaProject

qiskit_alt_path = os.path.dirname(os.path.abspath(__file__))


def new_project(calljulia="juliacall"):
    """Return a new `JuliaProject`.

    Use this if you want to use both `pyjulia` (`julia` module) and `juliacall`
    in a single session. For example, if are already using `pyjulia`, you can
    do `new_proj = new_project(calljulia="juliacall")` and then
    `new_proj.ensure_init()`.
    """
    return JuliaProject(
        name="qiskit_alt",
        package_path = qiskit_alt_path,
        version_spec = "^1.6", # Must be at least 1.6
        env_prefix = 'QISKIT_ALT_', # env variables prefixed with this may control JuliaProject
        registries = {"QuantumRegistry" : "https://github.com/Qiskit-Extensions/QuantumRegistry"},
        logging_level = logging.INFO, # or logging.WARN,
        console_logging=False,
        calljulia = calljulia
        )


project = new_project()



import os
from .julia_project import project
project.ensure_init()
from .pauli_operators import jlSparsePauliOp

from . import e_struct_python

Main = project.simple_import("Main")
QuantumOps = project.simple_import("QuantumOps")
QiskitQuantumInfo = project.simple_import("QiskitQuantumInfo")
ElectronicStructure = project.simple_import("ElectronicStructure")


def fermionic_hamiltonian(geometry, basis):
    """
    Given a molecular geometry specification and basis set, return
    the electronic Hamiltonian as a QuantumOps.FermiSum.

    The integrals are computed by pyscf. The geometry spec may be in either qiskit-nature
    format or qiskit_alt format.
    """
    mol_data = e_struct_python.MolecularData.from_specs(geometry=geometry, basis=basis)
    iop = ElectronicStructure.interaction_operator(mol_data.nuclear_repulsion,
                                                   mol_data.one_body_integrals,
                                                   mol_data.two_body_integrals)
    return QuantumOps.FermiSum(iop)


def jordan_wigner(fermi_op):
    """
    Compute the Jordan-Wigner transform of a QuantumOps.FermiSum representing a Fermionic Hamiltonian.
    Return a qiskit.SparsePauliOp.
    """
    pauli_op = QuantumOps.jordan_wigner(fermi_op)
    spop_jl = QiskitQuantumInfo.SparsePauliOp(pauli_op) # Convert to QiskitQuantumInfo.SparsePauliOp
    sparse_pauli = jlSparsePauliOp(spop_jl)  # Convert to qisit.quantum_info.SparsePauliOp
    return sparse_pauli



from collections import namedtuple
import functools
import pyscf
import numpy

"""
Pure-Python electronic structure code. This mostly provides an interface to PySCF.
"""


Atom = namedtuple("Atom", "species coords")


def atom(species, coords):
    if len(coords) != 3:
        raise ValueError(f"coords must be a three-tuple of floats. Got {coords}")
    return Atom(species, coords)


def atom_to_pyscf(atom):
    if not isinstance(atom, Atom):
        raise TypeError(f"Expect type `Atom`, got {type(atom)}'")
    return atom.species + " " + " ".join((str(coord) for coord in atom.coords))


Geometry = namedtuple("Geometry", "atoms")


def geometry(*atoms):
    if _is_nature_geometry(atoms[0]):
        return geometry_from_nature(atoms[0])
    if not all(isinstance(atom, Atom) for atom in atoms):
        raise ValueError(f"geometry: all arguments must be of type Atom")
    return Geometry(atoms)


def _is_nature_geometry(geom):
    return (isinstance(geom, list) and
            all((isinstance(g, list) and len(g) == 2 and
                 isinstance(g[0], str) and isinstance(g[1], list) and
                 len(g[1]) == 3) for g in geom)
            )


def geometry_from_nature(geom_list):
    if not _is_nature_geometry(geom_list):
        raise ValueError(f"{geom_list} is not a valid geometry specification")
    return Geometry([Atom(a[0], a[1]) for a in geom_list])


def geometry_to_pyscf(geometry):
    return ";".join((atom_to_pyscf(atom) for atom in geometry.atoms))


class MolecularSpec:
    def __init__(self, _geometry, multiplicity=None, charge=None, basis=None):
        self.geometry = geometry(_geometry)
        self.multiplicity = 1 if multiplicity is None else multiplicity
        self.charge = 0 if charge is None else charge
        self.basis = 'sto-3g' if basis is None else basis # probably should be required


    @property
    def spin(self):
        return self.multiplicity - 1


    def to_pyscf(self):
        pymol = pyscf.gto.Mole(atom=geometry_to_pyscf(self.geometry), basis=self.basis)
        pymol.spin = self.spin
        pymol.charge = self.charge
        pymol.symmetry = False
        pymol.build()
        return pymol


def one_electron_integrals(scf):
    mo_coeff = scf.mo_coeff
    h_core = scf.get_hcore()
    return functools.reduce(numpy.dot, (mo_coeff.T, h_core, mo_coeff))


def two_electron_integrals(pymol, scf):
    two_electron_compressed = pyscf.ao2mo.kernel(pymol, scf.mo_coeff)
    n_orbitals =  scf.mo_coeff.shape[1]
    symmetry_code = 1  # No permutation symmetry
    return pyscf.ao2mo.restore(symmetry_code, two_electron_compressed, n_orbitals)


class MolecularData:
    def __init__(self,
                 spec,
                 nuclear_repulsion,
                 one_body_integrals,
                 two_body_integrals
                 ):
        self.spec = spec
        self.nuclear_repulsion = nuclear_repulsion
        self.one_body_integrals = one_body_integrals
        self.two_body_integrals = two_body_integrals


    @classmethod
    def using_pyscf(cls, mol_spec : MolecularSpec):
        pymol = mol_spec.to_pyscf()
        if pymol.spin != 0:  # from OpenFermion
            scf = pyscf.scf.ROHF(pymol)
        else:
            scf = pyscf.scf.RHF(pymol)

        # Run Hartree-Fock
        verbose = scf.verbose
        scf.verbose = 0
        scf.run()
        scf.verbose = verbose
        one_e_ints = one_electron_integrals(scf)  # Compute one and two body integrals
        two_e_ints = two_electron_integrals(pymol, scf)
        nuclear_repulsion = float(pymol.energy_nuc()) # Compute constant energy, convert from numpy.float64 to float
        return MolecularData(mol_spec, nuclear_repulsion, one_e_ints, two_e_ints)


    @classmethod
    def from_specs(cls, geometry, multiplicity=None, charge=None, basis=None):
        mol_spec = MolecularSpec(geometry, multiplicity=multiplicity, charge=charge, basis=basis)
        return cls.using_pyscf(mol_spec)



__version__ = "0.1.7"

from .julia_project import project

# QuantumOps = project.simple_import("QuantumOps")
# from julia import QuantumOps, QiskitQuantumInfo, ElectronicStructure
# from .pauli_operators import jlPauli, jlSparsePauliOp, PauliSum_to_SparsePauliOp
# from .electronic_structure import Geometry, jordan_wigner, fermionic_hamiltonian



import os

from qiskit.quantum_info import Pauli, SparsePauliOp, PauliList
from .julia_project import project
project.ensure_init()

Main = project.simple_import("Main")
QuantumOps = project.simple_import("QuantumOps")
QiskitQuantumInfo = project.simple_import("QiskitQuantumInfo")

if project._calljulia_name == 'juliacall':
    Main.include(os.path.join(os.path.dirname(os.path.abspath(__file__)), 'juliacall_util.jl'))


def jlPauli(data):
    """
    Convert a QiskitQuantumInfo.Pauli to a qiskit.quantum_info.Pauli
    """
    if isinstance(data, str):
        data = QiskitQuantumInfo.Pauli(data)
    return Pauli((data.x, data.z, data.phase))


def jlPauliList(pauli_list):
    """
    Convert a QiskitQuantumInfo.PauliList to a qiskit.quantum_info.PauliList
    """
    return PauliList.from_symplectic(pauli_list.z, pauli_list.x)


def jlSparsePauliOp(sp):
    pl = PauliList.from_symplectic(sp.pauli_list.z, sp.pauli_list.x)
    return SparsePauliOp(pl, Main.map(Main.float, sp.coeffs))


def PauliSum_to_SparsePauliOp(ps):
    """
    Convert a QuantumOps.PauliSum to a qiskit.quantum_info.SparsePauliOp
    """
    spop_jl = QiskitQuantumInfo.SparsePauliOp(ps) # Convert to QiskitQuantumInfo.SparsePauliOp
    spop = jlSparsePauliOp(spop_jl)  # Convert to qisit.quantum_info.SparsePauliOp
    return spop
